package metaobj;

import meta.AnnotationArgumentsKind;
import meta.AttachedDeclarationKind;
import meta.CyanMetaobjectWithAt;
import meta.ICheckOverride_afsa;
import meta.ICompiler_dpa;
import meta.ICompiler_dsa;
import meta.IParseWithCyanCompiler_dpa;
import meta.Token;
import meta.WrASTVisitor;
import meta.WrEnv;
import meta.WrMethodDec;
import meta.WrStatementBreak;
import meta.WrStatementIf;
import meta.WrStatementWhile;

/**
 * The annotation of this metaobject should be attached to a method.
 * When it is overridden in a subprototype, the metaobject checks
 * the code of the subprototype method based on the code of the attached
 * DSL. The annotation accepts an attached DSL with the syntax:<br>
 * <code><br>
 *     Start :: = Stat { "," Stat }<br>
 *     Stat :: = "if:" N | "while:" N | "repeat:" N <br>
 *                 | no break<br>
 *  </code><br>
 *
 * N is an integer number. The semantics is: if the DSL code has
 * stat: N, then  'stat' can only appear in the code N times.
 * If 'no break' is present, the code should have no 'break'
 * statement. Then, if the annotation is<br>
 * </code<br>
         {@literal @}three<*<br>
         	if: 2, <br>
         	no break<br>
         *><br>
   <code>
   the overridden method can have any number of while and repeat statements but
   only two if statements. And no breaks.


   @author jose
 */
public class CyanMetaobjectThree extends CyanMetaobjectWithAt
    implements IParseWithCyanCompiler_dpa, ICheckOverride_afsa {

    public CyanMetaobjectThree() {
        super("three", AnnotationArgumentsKind.ZeroOrMoreParameters,
                new AttachedDeclarationKind[] {
                AttachedDeclarationKind.METHOD_DEC
        });
    }

    private int numberIFsAllowed = -1;
    private int numberWHILEsAllowed = -1;
    private boolean allowBreaks = true;


    @Override
	public boolean shouldTakeText() { return true; }

	@Override
	public void dpa_parse(ICompiler_dpa compiler_dpa) {
		compiler_dpa.next();
		while ( compiler_dpa.getSymbol().token != Token.EOLO ) {

			switch ( compiler_dpa.getSymbol().token ) {
			case IDENT:
				String id = compiler_dpa.getSymbol().getSymbolString();
				if ( !id.equalsIgnoreCase("no") ) {
					compiler_dpa.error(compiler_dpa.getSymbol(),
							"'no' expected");
				}
				compiler_dpa.next();
				if ( compiler_dpa.getSymbol().token != Token.BREAK ) {
					compiler_dpa.error(compiler_dpa.getSymbol(),
							"'break' expected");
				}
//				id = compiler_dpa.getSymbol().getSymbolString();
//				if ( !id.equalsIgnoreCase("break") ) {
//					compiler_dpa.error(compiler_dpa.getSymbol(),
//							"'break' expected");
//				}
				allowBreaks = false;
				compiler_dpa.next();
				break;
			case IDENTCOLON:
				switch ( compiler_dpa.getSymbol().getSymbolString() ) {
				case "if:" :
					compiler_dpa.next();
					if ( compiler_dpa.getSymbol().token != Token.INTLITERAL ) {
						compiler_dpa.error(compiler_dpa.getSymbol(),
								"Literal int expected");
					}
					this.numberIFsAllowed = Integer.parseInt(
							compiler_dpa.getSymbol().getSymbolString());
					compiler_dpa.next();
					break;
				case "while:" :
					compiler_dpa.next();
					if ( compiler_dpa.getSymbol().token != Token.INTLITERAL ) {
						compiler_dpa.error(compiler_dpa.getSymbol(),
								"Literal int expected");
					}
					this.numberWHILEsAllowed = Integer.parseInt(
							compiler_dpa.getSymbol().getSymbolString());
					compiler_dpa.next();
					break;
				default:
					compiler_dpa.error(compiler_dpa.getSymbol(),
							"'if:' or 'while:' expected. Found '" +
					   compiler_dpa.getSymbol().getSymbolString() + "'");
				}
				break;
			default:
				compiler_dpa.error(compiler_dpa.getSymbol(),
						"'if:', 'while:' or 'no' expected");
			}
			if ( compiler_dpa.getSymbol().token != Token.COMMA ) {
				break;
			}
			else {
				compiler_dpa.next();
			}
		}
		if ( compiler_dpa.getSymbol().token != Token.EOLO ) {
			compiler_dpa.error(compiler_dpa.getSymbol(), "End of file expected");
		}
	}



	@Override
	public void afsa_checkOverride(ICompiler_dsa compiler_dsa,
			WrMethodDec method) {
		WrEnv env = compiler_dsa.getEnv();
		StatCountASTVisitor visitor = new StatCountASTVisitor();
		method.accept(visitor, env);
		if ( this.numberIFsAllowed >= 0 && visitor.ifCount > this.numberIFsAllowed ) {
			this.addError(method.getFirstSymbol(env),
					"This method has " + visitor.ifCount + " 'if' statements."
							+ " The number allowed is " + this.numberIFsAllowed);
		}

		if ( this.numberWHILEsAllowed >= 0 && visitor.whileCount > this.numberWHILEsAllowed) {
			this.addError(method.getFirstSymbol(env),
					"This method has " + visitor.whileCount + " 'while' statements."
							+ " The number allowed is " + this.numberWHILEsAllowed);
		}

		if ( ! this.allowBreaks && visitor.breakCount > 0) {
			this.addError(method.getFirstSymbol(env),
					"This method has a 'break' statement but none is allowed");
		}
	}

}


class StatCountASTVisitor extends WrASTVisitor {
    public CyanMetaobjectWithAt metaobject;
    public int ifCount = 0;
    public int whileCount = 0;
    public int breakCount = 0;

    @Override
    public void visit(WrStatementWhile node, WrEnv env) {
        ++whileCount;
    }

    @Override
    public void visit(WrStatementIf node, WrEnv env) {
        ++ifCount;
    }

    @Override
    public void visit(WrStatementBreak node, WrEnv env) {
        ++breakCount;
    }

}